(*
  Copyright 2010 Benot Barbot

  This file is part of ReedsShepp.

  ReedsShepp is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  ReedsShepp is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with ReedsShepp.  If not, see <http://www.gnu.org/licenses/>.
  
*)

open Unix;;
open Reedandshepp;;

type ball = ReedsAndShepp | Dubin | DubinAcc;;
type dessin = Full | Mpipi | PiM | PiP;;

let resolx= 800;;
let resoly= 800;;
let anglex=ref 0.
and angley=ref 0.
and anglez=ref 0.
and vitx = ref 0.
and vity = ref 0.
and vitz = ref 0.
and echelle =ref 1.5
and rayon = ref 1.
and retrace = ref true;
and color = ref true;
and ball = ref ReedsAndShepp;
and modulo = ref true;
and repert = ref true;
and sym = ref true;
and help = ref 1;
and zoom = ref true;
and restr = ref Full;
and coupe = ref false;
and ta = Array.make 12 true;;

(*if Array.length Sys.argv >1 then rayon := float_of_string Sys.argv.(1);;*)

let grey = (0.3,0.3,0.3);;

let scalaire (x1,y1,z1) (x2,y2,z2) =
  x1 *. x2 +. y1*.y2 +. z1*. z2;;

let vectoriel (x1,y1,z1) (x2,y2,z2) =
    (y1*.z2 -. y2*.z1 , z1*.x2 -.x1 *.z2 , x1 *.y2 -. x2 *.y1);;

let diff (x1,y1,z1) (x2,y2,z2) =
  (x1 -. x2 , y1 -. y2 , z1 -. z2);;

let moins (x1,y1,z1) =
  (-.x1 , -.y1 , -.z1);;

let normalise (x,y,z) =
  let r =  sqrt (x *.x +. y*.y +. z*.z) in
    (x/.r, y/.r, z/.r);;

let conddes z = function
    Full -> true
  | Mpipi -> z <= pi && z >= -.pi
  | PiM -> z <= -.pi && z >= -.3.*.pi
  | PiP -> z >= pi && z <= 3.*.pi;;

let pred_vect1 = ref (0.,0.,0.)
and pred_vect2 = ref (0.,0.,0.);;
let dessin v b =
  let _,_,z =v in
    if (conddes z !restr) then begin
	  if b then
	    GlDraw.normal3 (normalise ((vectoriel (diff !pred_vect2 v) (diff !pred_vect1 v))))
	  else 
	    GlDraw.normal3 (normalise ((vectoriel (diff !pred_vect1 v) (diff !pred_vect2 v))));
      GlDraw.vertex3 v;
      pred_vect2 := !pred_vect1;
      pred_vect1 := v;
    end;;

let cone x y z =
  let orig = (z,0.,0.) in
    GlDraw.begins `lines;
    GlDraw.vertex3 (0.,0.,0.);
    GlDraw.vertex3 orig;
    GlDraw.ends ();
    GlDraw.begins `triangle_fan;
    GlDraw.vertex3 orig;
    for i =0 to 20 do
      let k = (float_of_int i)*. Reedandshepp.pi /. 10. in
      let point = (z-.x,y *. (cos k),y*. (sin k)) 
      and point2 = (0., -.(sin k), (cos k) ) in
	GlDraw.normal3 (normalise ((vectoriel (diff point orig) point2)));
	GlDraw.vertex3 point;
    done;
  GlDraw.ends ();
  GlDraw.begins `triangle_fan;
  GlDraw.normal3 (-.1.,0.,0.);
  GlDraw.vertex3 (z-.x,0.,0.);
  for i =20 downto 0 do
    let k = (float_of_int i)*. Reedandshepp.pi /. 10. in
      GlDraw.vertex3 (z-.x,y *. (cos k),y*. (sin k));
  done;
  GlDraw.ends ();;

let set_proj x y = 
  GlDraw.viewport ~x:0 ~y:0 ~w:x ~h:y;
  GlMat.mode `projection;
  GlMat.load_identity ();
  GluMat.perspective ~fovy:45.0 ~aspect:((float_of_int x) /. (float_of_int y)) ~z:(0.1, 1000.);
  GluMat.look_at ~eye:(10., 0., 0.) ~center:(0., 0., 0.) ~up:(0., 0., 1.);;

let setup () =
  set_proj resolx resoly;
  GlClear.color (0.6, 1., 0.8);
  let objlist = GlList.gen_lists ~len:4 in
    GlList.begins (GlList.nth objlist ~pos:2) ~mode:`compile;
    GlDraw.color (1.,0.,0.);
    cone 0.2 0.1 Reedandshepp.pi;
    GlMat.push ();
    GlMat.rotate ~angle:90. ~z:1. ();
    GlDraw.color (0.,1.,0.);
    cone 0.2 0.1 Reedandshepp.pi;
    GlMat.pop ();
    GlMat.push ();
    GlMat.rotate ~angle:(-90.) ~y:1. ();
    GlDraw.color (0.,0.,1.);
    cone 0.2 0.1 Reedandshepp.pi;
    GlMat.pop ();
    GlList.ends ();

    (*GlList.begins (GlList.nth objlist ~pos:0) ~mode:`compile;
    boule 100 dessin !color !rayon;
    GlList.ends ();*)
    Gl.enable `depth_test;
    (*GlDraw.cull_face `back; 
    GlDraw.front_face `ccw;
    Gl.enable `cull_face;*)
    (*Gl.enable `alpha_test;*)
    Gl.enable `lighting;
    Gl.enable `light0; 
    GlLight.light ~num:0 (`position (10., 0., 0., 2.));
    Gl.enable `color_material;
    objlist;;

let draw_string x y s =
  GlPix.raster_pos ~x:(5.) ~y:x ~z:y ();
  for i =0 to String.length s -1 do
    Glut.bitmapCharacter ~font:Glut.BITMAP_HELVETICA_18 ~c:(int_of_char s.[i]);
  done;;

let render objlist () =
  if !retrace then begin
    if !coupe then begin
      GlList.begins (GlList.nth objlist ~pos:3) ~mode:`compile;
      restr := PiP;
      begin
	match !ball with
	    ReedsAndShepp ->
	      boule 100 dessin !color !rayon ta
	  | Dubin ->
	      bouleD 100 dessin !color !rayon ta
	  | DubinAcc ->
	      bouleDR 150 dessin !color !rayon ta;
      end;
      GlMat.scale3 (1.,-1.,-1.);
      restr := PiM;
      begin
	match !ball with
	    ReedsAndShepp ->
	      boule 100 dessin !color !rayon ta
	  | Dubin ->
	      bouleD 100 dessin !color !rayon ta
	  | DubinAcc ->
	      bouleDR 150 dessin !color !rayon ta;
      end;
      
      GlList.ends ();
      GlList.begins (GlList.nth objlist ~pos:1) ~mode:`compile;
      restr := Mpipi;
      begin
	match !ball with
	    ReedsAndShepp ->
	      boule 100 dessin !color !rayon ta
	  | Dubin ->
	      bouleD 100 dessin !color !rayon ta
	  | DubinAcc ->
	      bouleDR 150 dessin !color !rayon ta;
      end;
      
      GlMat.push ();
      GlMat.translate ~z:(-.pi2) ();
      GlList.call (GlList.nth objlist ~pos:3);
      GlMat.pop ();

      GlList.ends ();
    end else begin
      if !ball = ReedsAndShepp 
      then restr := Mpipi
      else restr := Full;
      GlList.begins (GlList.nth objlist ~pos:1) ~mode:`compile;
      begin
	match !ball with
	    ReedsAndShepp ->
		boule 100 dessin !color !rayon ta
	  | Dubin ->
	      bouleD 100 dessin !color !rayon ta
	  | DubinAcc ->
	      bouleDR 150 dessin !color !rayon ta;
      end;
      GlList.ends ();
    end;
    GlList.begins (GlList.nth objlist ~pos:0) ~mode:`compile;
    GlList.call (GlList.nth objlist ~pos:1);
    if !sym then begin
      GlMat.push ();
      GlMat.scale3 (1.,-1.,-1.);
      GlList.call (GlList.nth objlist ~pos:1);
      GlMat.pop ();
      if !ball = ReedsAndShepp then begin
	GlMat.push ();
	GlMat.scale3 (-1.,-1.,1.);
	GlList.call (GlList.nth objlist ~pos:1);
	GlMat.pop ();
	GlMat.push ();
	GlMat.scale3 (-1.,1.,-1.);
	GlList.call (GlList.nth objlist ~pos:1);
	GlMat.pop ();
      end;
    end;
    GlList.ends ();
    retrace := false;
  end;
  
  GlMat.mode `modelview;
  GlMat.load_identity ();
  
  GlMat.push ();
  
  if !zoom then
    let echelle2 = (!echelle /. !rayon ) in
      GlMat.scale ~x: echelle2 ~y: echelle2 ~z: echelle2  ()
  else GlMat.scale ~x: !echelle ~y: !echelle ~z: !echelle  ();
      
      GlMat.rotate ~angle:!anglex ~x:1. ();
      GlMat.rotate ~angle:!angley ~y:1. ();
      GlMat.rotate ~angle:!anglez ~z:1. ();
  (*GlMat.rotate3 (glpoint_of_matrix angle);*)

  GlClear.clear [`color; `depth ];  

  if !repert then
    GlList.call (GlList.nth objlist ~pos:2);
  GlList.call (GlList.nth objlist ~pos:0);
  
  if !modulo then begin
    GlMat.push ();
    GlMat.translate ~z:(-.3.*.pi2) ();
    for i = -3 to 3 do 
      GlList.call (GlList.nth objlist ~pos:0);
      GlMat.translate ~z:pi2 ();
    done;
    GlMat.pop ();
  end else
    GlList.call (GlList.nth objlist ~pos:0);
  
  GlMat.pop ();

  if !help =1 then begin
    draw_string (-.2.) (1.95) (Printf.sprintf "Rayon de la boule: %f = %f pi" !rayon (!rayon /. pi));
    draw_string (-.2.) (1.85) "Tapez h pour l'aide";
  end;
  if !help =2 then begin
    draw_string (-.2.) (1.95) "Esc pour quitter";
    draw_string (-.2.) (1.85) "h pour l'aide";
    draw_string (-.2.) (1.75) "y pour augmenter le rayon";
    draw_string (-.2.) (1.65) "u pour diminuier le rayon";
    draw_string (-.2.) (1.55) "r pour zoomer";
    draw_string (-.2.) (1.45) "t pour dezoomer";
    draw_string (-.2.) (1.35) "z,s q,d w,c pour tourner";
    draw_string (-.2.) (1.25) "f pour le mode pleine ecran";
    draw_string (-.2.) (1.15) "option par clique droit sur la boulle";
  end;

  Gl.flush ();
  Glut.swapBuffers ();;


let keyboard ~key ~x ~y = match char_of_int key with
    '\027'->begin
      exit 0
    end
  | 'h'   ->help:= (!help +1) mod 3
  | 'z'   ->angley:=!angley-.2.
  | 's'   ->angley:=!angley+.2.
  | 'q'   ->anglez:=!anglez+.2.
  | 'd'   ->anglez:=!anglez-.2.
  | 'w'   ->anglex:=!anglex-.2.
  | 'c'   ->anglex:=!anglex+.2.
  | 'r'   ->echelle:= !echelle*.1.1
  | 't'   ->echelle:= !echelle/.1.1
  | 'y'   ->rayon := !rayon*.1.1;
      retrace := true;
  | 'u'   ->rayon := !rayon/.1.1;
      retrace := true;
  | 'f'   ->Glut.fullScreen ()
  | _->();;

let xmouse = ref 0;;
let ymouse = ref 0;;

let mouse ~button ~state ~x ~y =
    xmouse := x;
    ymouse := y;;
  
let mouvmouse ~x ~y  =
  anglez:=!anglez +. float_of_int (x - !xmouse);
  angley:=!angley +. float_of_int (y - !ymouse);
  xmouse := x;
  ymouse := y;;

let radmenu i s = 
  if ta.(i) then
    Glut.changeToMenuEntry ~item:(i+1) ~label:("+"^s) ~value:(i+11)
  else Glut.changeToMenuEntry ~item:(i+1) ~label:("-"^s) ~value:(i+11);;

let changemenu1 ()=
  radmenu 0 " C S C";
  radmenu 1 " C S C";
    radmenu 2 " C | C | C";
    radmenu 3 " C | C C";
    radmenu 4 " C C | C";
    radmenu 5 " C Cu | Cu C";
    radmenu 6 " C | Cu Cu | C";
    radmenu 7 " C | Cpi/2 S C";
    radmenu 8 " C S Cpi/2 | C";
    radmenu 9 " C | Cpi/2 S C";
    radmenu 10 " C S Cpi/2 | C";
    radmenu 11 " C | Cpi/2 S Cpi/2 | C";;
let changemenu2 ()=
  radmenu 0 " C C C";
  radmenu 1 " C S C";
    radmenu 2 " C S C";
    radmenu 3 "";
    radmenu 4 "";
    radmenu 5 "";
    radmenu 6 "";
    radmenu 7 "";
    radmenu 8 "";
    radmenu 9 "";
    radmenu 10 "";
    radmenu 11 "";;
let changemenu () =
  if !ball = ReedsAndShepp then changemenu1 ()
  else changemenu2 ();;

let menufonct drawmenu ~value:x = 
  retrace := true;
  begin
    match x with
	1 -> color := true
      | 2 -> color := false
      | 3 -> ball := ReedsAndShepp
      | 4 -> ball := Dubin
      | 27 -> ball := DubinAcc
      | 5 -> modulo := true
      | 6 -> modulo := false
      | 7 -> repert := true
      | 8 -> repert := false
      | 9 -> sym := true
      | 10 -> sym := false
      | 23 -> Array.fill ta 0 12 true
      | 24 -> Array.fill ta 0 12 false
      | 25 -> zoom := true
      | 26 -> zoom := false
      | 28 -> coupe := true;
	  modulo := false
      | 29 -> coupe := false;
	  modulo := true
      | x ->ta.(x-11) <- not ta.(x-11);
  end;
  Glut.setMenu drawmenu;
  changemenu ();;
  

let reshape ~w:x ~h:y =
  set_proj x y;; 

let _ =
  let _ = Glut.init Sys.argv in
  Glut.initDisplayMode ~depth:true ~double_buffer:true ();
  Glut.initWindowSize ~w: resolx ~h:resoly;
  ignore(Glut.createWindow ~title:"Boule de Reeds & Shepp");
  let objlist = setup () in
  Glut.displayFunc ~cb:(render objlist);
  Glut.idleFunc ~cb:(Some Glut.postRedisplay);
  Glut.keyboardFunc ~cb:keyboard;
  Glut.mouseFunc ~cb:mouse;
  Glut.motionFunc ~cb:mouvmouse;
  Glut.reshapeFunc ~cb:reshape;
  let a = ref 0 in
  let rec drawmenu = Glut.createMenu ~cb:(menufonct !a) in
    Glut.addMenuEntry ~label:"+ C S C" ~value:11 ;
    Glut.addMenuEntry ~label:"+ C S C" ~value:12 ;
    Glut.addMenuEntry ~label:"+ C | C | C" ~value:13 ;
    Glut.addMenuEntry ~label:"+ C | C C" ~value:14 ;
    Glut.addMenuEntry ~label:"+ C C | C" ~value:15 ;
    Glut.addMenuEntry ~label:"+ C Cu | Cu C" ~value:16 ;
    Glut.addMenuEntry ~label:"+ C | Cu Cu | C" ~value:17 ;
    Glut.addMenuEntry ~label:"+ C | Cpi/2 S C" ~value:18 ;
    Glut.addMenuEntry ~label:"+ C S Cpi/2 | C" ~value:19 ;
    Glut.addMenuEntry ~label:"+ C | Cpi/2 S C" ~value:20 ;
    Glut.addMenuEntry ~label:"+ C S Cpi/2 | C" ~value:21 ;
    Glut.addMenuEntry ~label:"+ C | Cpi/2 S Cpi/2 | C" ~value:22 ;
    Glut.addMenuEntry ~label:"Toutes" ~value:23 ;
    Glut.addMenuEntry ~label:"Aucunes" ~value:24 ;
    a:=drawmenu;
    let menufonct2 = menufonct drawmenu in
    let colormenu = Glut.createMenu ~cb:menufonct2 in
      Glut.addMenuEntry ~label:"Polychrone" ~value:1 ;
      Glut.addMenuEntry ~label:"Monochrone" ~value:2 ;
      let ballmenu = Glut.createMenu ~cb:menufonct2 in
	Glut.addMenuEntry ~label:"Reeds & Shepp" ~value:3 ;
	Glut.addMenuEntry ~label:"Dubins" ~value:4 ;
	Glut.addMenuEntry ~label:"Dubins Accessible" ~value:27 ;
	let modulomenu = Glut.createMenu ~cb:menufonct2 in
	  Glut.addMenuEntry ~label:"Oui" ~value:5 ;
	  Glut.addMenuEntry ~label:"Non" ~value:6 ;
	  let repertmenu = Glut.createMenu ~cb:menufonct2 in
	    Glut.addMenuEntry ~label:"Oui" ~value:7 ;
	    Glut.addMenuEntry ~label:"Non" ~value:8 ;
	    let symmenu = Glut.createMenu ~cb:menufonct2 in
	      Glut.addMenuEntry ~label:"Oui" ~value:9 ;
	      Glut.addMenuEntry ~label:"Non" ~value:10 ;
	      let zoommenu = Glut.createMenu ~cb:menufonct2 in
		Glut.addMenuEntry ~label:"Oui" ~value:25 ;
		Glut.addMenuEntry ~label:"Non" ~value:26 ;
		let restrmenu = Glut.createMenu ~cb:menufonct2 in
		  Glut.addMenuEntry ~label:"Oui" ~value:28 ;
		  Glut.addMenuEntry ~label:"Non" ~value:29 ;
		let _ = Glut.createMenu ~cb:menufonct2 in
		  Glut.addSubMenu ~label:"Couleur" ~submenu:colormenu ;
		  Glut.addSubMenu ~label:"Boule" ~submenu:ballmenu ;
		  Glut.addSubMenu ~label:"Modulo 2 pi" ~submenu:modulomenu ;
		  Glut.addSubMenu ~label:"Repere" ~submenu:repertmenu ;
		  Glut.addSubMenu ~label:"Zoom auto" ~submenu:zoommenu ;
		  Glut.addSubMenu ~label:"Chemin" ~submenu:drawmenu ;
		  Glut.addSubMenu ~label:"Symetrie" ~submenu:symmenu ;
		  Glut.addSubMenu ~label:"Restriction" ~submenu:restrmenu ;
		  Glut.attachMenu ~button:Glut.RIGHT_BUTTON ;
		  Glut.setMenu drawmenu;
		  Glut.mainLoop ();;

