(*
  Copyright 2010 Benot Barbot

  This file is part of ReedsShepp.

  ReedsShepp is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  ReedsShepp is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with ReedsShepp.  If not, see <http://www.gnu.org/licenses/>.
  
*)

let pi = 4.*. atan 1.;;
let pi2 = 2.*.pi;;

type manoeuvre = Sp | Sm | Lp | Lm |Rp | Rm;;  

let rec calcmanoeuvre (x,y,phi) t = function
    Lp -> (x +. (cos phi) *. (sin t) -. (sin phi) *. (1. -. cos t) ,
	   y +. (sin phi) *. (sin t) +. (cos phi) *. (1. -. cos t) , phi +.t)
  | Lm -> calcmanoeuvre (x,y,phi) (-.t) Lp 
  | Rp -> (x +. (cos phi) *. (sin t) +. (sin phi) *. (1. -. cos t) ,
	   y +. (sin phi) *. (sin t) -. (cos phi) *. (1. -. cos t) , phi -.t)
  | Rm -> calcmanoeuvre (x,y,phi) (-.t) Rp
  | Sp -> (x +. t *. (cos phi) , y +. t*. (sin phi) , phi)
  | Sm -> calcmanoeuvre (x,y,phi) (-.t) Sp;;

let sym1 = List.map (function
    Sp -> Sm | Sm -> Sp
  | Lp -> Lm | Lm -> Lp
  | Rp -> Rm | Rm -> Rp);;
let sym2 = List.map (function 
    Lp -> Rp | Lm -> Rm
  | Rp -> Lp | Rm -> Lm
  | x -> x);;
let sym3 = fun x -> sym1 (sym2 x);;

let coeftuv i def r b =
  if b then
    (float_of_int i) *. r /. def
  else  r -. (float_of_int i) *. r /. def;;

let dessinesurf def dessin r b col l coef =
  let raytab = Array.make (def+1) (0.,0.,0.) in
  let def2 = float_of_int def in
      GlDraw.color col;
    for i = 0 to def do
      let t = coeftuv i def2 r (not b) in
      GlDraw.begins `triangle_strip;
      for j = 0 to def do
	let u = coeftuv j def2 (r-.t) true in
	let v = r -. t -. u in
	let rayon = List.fold_left2 calcmanoeuvre  (0.,0.,0.) (coef t u v) l in
	  if i > 0 then begin
	    dessin raytab.(j) true;
	    dessin rayon false;
	  end;
	  raytab.(j) <- rayon;
      done;
      GlDraw.ends ();
    done;;

let dessinesurf2 def dessin r b col l coef =
  let raytab = Array.make (def+1) (0.,0.,0.) in
  let def2 = float_of_int def in
      GlDraw.color col;
    for i = 0 to def do
      let t = coeftuv i def2 (1.) (not b) in
      GlDraw.begins `triangle_strip;
      for j = 0 to def do
	let u = coeftuv j def2 (1.) true in
	let rayon = List.fold_left2 calcmanoeuvre  (0.,0.,0.) (coef t u 0.) l in
	  if i > 0 then begin
	    dessin raytab.(j) true;
	    dessin rayon false;
	  end;
	  raytab.(j) <- rayon;
      done;
      GlDraw.ends ();
    done;;
  
let dessinesurfsym def dessin r b col l coef =
  dessinesurf def dessin r b col l coef;
  dessinesurf def dessin r b col (sym1 l) coef;
  dessinesurf def dessin r b col (sym2 l) coef;
  dessinesurf def dessin r b col (sym3 l) coef;;

let boule2 def dessin r =
  dessinesurf def dessin (r-.pi/.2.) true (1.,0.,0.) [Sp;Lp;Sp] (fun t u v -> [u;pi/.2.;v]) ;;

let boule def dessin col r ta=
  if col then begin
    if ta.(0) then dessinesurf def dessin r true (1.,0.,0.) [Lp;Sp;Lp] (fun t u v->[t;u;v]);
    if ta.(1) then dessinesurf def dessin r false (1.,1.,0.) [Lp;Sp;Rp] (fun t u v->[t;u;v]);
    if ta.(3) then dessinesurf def dessin r false (0.,1.,0.) [Lp;Rm;Lm] (fun t u v->[t;u;v]);
    if ta.(4) then dessinesurf def dessin r false (0.,1.,1.) [Lm;Rm;Lp] (fun t u v->[t;u;v]);
    if ta.(5) then dessinesurf def dessin r false (0.,0.,1.) [Lp;Rp;Lm;Rm] (fun t u v ->[t;u/.2.;u/.2.;v]);
    if ta.(6) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Rm;Lm;Rp] (fun t u v -> [t;u/.2.;u/.2.;v]);
    if r >= pi /.2. then begin
      if ta.(7) then dessinesurf def dessin (r-.pi/.2.) false (1.,0.5,0.) [Lp;Rm;Sm;Lm] (fun t u v->[t;pi/.2.;u;v]);
      if ta.(8) then dessinesurf def dessin (r-.pi/.2.) false (0.5,1.,0.) [Lm;Sm;Rm;Lp] (fun t u v->[t;u;pi/.2.;v]);
      if ta.(9) then dessinesurf def dessin (r-.pi/.2.) true (0.,0.5,0.1) [Lp;Rm;Sm;Rm] (fun t u v->[t;pi/.2.;u;v]);
      if ta.(10) then dessinesurf def dessin (r-.pi/.2.) true (0.,1.,0.5) [Rm;Sm;Rm;Lp] (fun t u v->[t;u;pi/.2.;v]);
    end;
    if r >= pi then begin
      if ta.(11) then dessinesurf def dessin (r-.pi) false (0.5,0.,0.5) [Lp;Rm;Sm;Lm;Rp] (fun t u v->[t;pi/.2.;u;pi/.2.;v]);
    end
    else if ta.(2) then dessinesurf def dessin r true (0.5,1.,0.5) [Lp;Rm;Lp] (fun t u v->[t;u;v]);
      
  end else begin
    if ta.(0) then dessinesurf def dessin r true (0.5,0.5,0.5) [Lp;Sp;Lp] (fun t u v->[t;u;v]);
    if ta.(1) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Sp;Rp] (fun t u v->[t;u;v]);
    if ta.(3) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Rm;Lm] (fun t u v->[t;u;v]);
    if ta.(4) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lm;Rm;Lp] (fun t u v->[t;u;v]);
    if ta.(5) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Rp;Lm;Rm] (fun t u v ->[t;u/.2.;u/.2.;v]);
    if ta.(6) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Rm;Lm;Rp] (fun t u v -> [t;u/.2.;u/.2.;v]);
    if r >= pi /.2. then begin
      if ta.(7) then dessinesurf def dessin (r-.pi/.2.) false (0.5,0.5,0.5) [Lp;Rm;Sm;Lm] (fun t u v->[t;pi/.2.;u;v]);
      if ta.(8) then dessinesurf def dessin (r-.pi/.2.) false (0.5,0.5,0.5) [Lm;Sm;Rm;Lp] (fun t u v->[t;u;pi/.2.;v]);
      if ta.(9) then dessinesurf def dessin (r-.pi/.2.) true (0.5,0.5,0.5) [Lp;Rm;Sm;Rm] (fun t u v->[t;pi/.2.;u;v]);
      if ta.(10) then dessinesurf def dessin (r-.pi/.2.) true (0.5,0.5,0.5) [Rm;Sm;Rm;Lp] (fun t u v->[t;u;pi/.2.;v]);
    end;
    if r >= pi then begin
      if ta.(11) then dessinesurf def dessin (r-.pi) false (0.5,0.5,0.5) [Lp;Rm;Sm;Lm;Rp] (fun t u v->[t;pi/.2.;u;pi/.2.;v])
    end
    else if ta.(2) then dessinesurf def dessin r true (0.5,0.5,0.5) [Lp;Rm;Lp] (fun t u v->[t;u;v]);
    
  end;;

let bouleD def dessin col r ta =
  if col then
    begin
      if ta.(0) && r >= pi then begin
	dessinesurf2 def dessin r true (1.,0.,0.) [Rp;Lp;Rp] (
	  fun t u v->
	    let b = pi +. t*. (r -. pi) in
	    let a = min (u *. (b-. pi)) (r-.b) in
	      [a;b;r-.b-.a]);
	dessinesurf2 def dessin r false (1.,0.,0.) [Rp;Lp;Rp] (
	  fun t u v->
	    let b = pi +. t*. (r -. pi) in
	    let a = min (u *. (b-. pi)) (r-.b) in
	      [r-.b-.a;b;a]);
      end;
      if ta.(1) then dessinesurf def dessin r true (0.,1.,0.) [Lp;Sp;Lp] (fun t u v->[t;u;v]);
      if ta.(2) then dessinesurf def dessin r false (0.,0.,1.) [Lp;Sp;Rp] (fun t u v->[t;u;v]);
    end
  else begin
    if ta.(0) && r >= pi then begin
	dessinesurf2 def dessin r true (0.5,0.5,0.5) [Rp;Lp;Rp] (
	  fun t u v->
	    let b = pi +. t*. (r -. pi) in
	    let a = min (u *. (b-. pi)) (r-.b) in
	      [a;b;r-.b-.a]);
	dessinesurf2 def dessin r false (0.5,0.5,0.5) [Rp;Lp;Rp] (
	  fun t u v->
	    let b = pi +. t*. (r -. pi) in
	    let a = min (u *. (b-. pi)) (r-.b) in
	      [r-.b-.a;b;a]);
      end;
    if ta.(1) then dessinesurf def dessin r true (0.5,0.5,0.5) [Lp;Sp;Lp] (fun t u v->[t;u;v]);
    if ta.(2) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Sp;Rp] (fun t u v->[t;u;v]);
  end;;
    
let bouleDR def dessin col r ta =
  if col then
    begin
      if ta.(0) then begin
	dessinesurf def dessin r false (1.,0.,0.) [Rp;Lp;Rp] (fun t u v->[t;u;v]);
	dessinesurf def dessin r false (1.,0.5,0.5) [Rp;Lp;Rp] (fun t u v->[0.;u;v]);
	dessinesurf def dessin r false (1.,0.5,0.5) [Rp;Lp;Rp] (fun t u v->[t;u;0.]);
      end;
      if ta.(1) then begin
	dessinesurf def dessin r true (0.,1.,0.) [Lp;Sp;Lp] (fun t u v->[t;u;v]);
	dessinesurf def dessin r false (0.5,1.,0.5) [Lp;Sp;Lp] (fun t u v->[0.;u;v]);
	dessinesurf def dessin r false (0.5,1.,0.5) [Lp;Sp;Lp] (fun t u v->[t;u;0.]);
      end;
      if ta.(2) then begin
	dessinesurf def dessin r false (0.,0.,1.) [Lp;Sp;Rp] (fun t u v->[t;u;v]);
	dessinesurf def dessin r true (0.5,0.5,1.) [Lp;Sp;Rp] (fun t u v->[0.;u;v]);
	dessinesurf def dessin r true (0.5,0.5,1.) [Lp;Sp;Rp] (fun t u v->[t;0.;v]);
      end;
    end else begin
      if ta.(0) then dessinesurf def dessin r false (0.5,0.5,0.5) [Rp;Lp;Rp] (fun t u v->[t;u;v]);
      if ta.(0) then dessinesurf def dessin r false (0.5,0.5,0.5) [Rp;Lp;Rp] (fun t u v->[0.;u;v]);
      if ta.(0) then dessinesurf def dessin r false (0.5,0.5,0.5) [Rp;Lp;Rp] (fun t u v->[t;u;0.]);

      if ta.(1) then dessinesurf def dessin r true (0.5,0.5,0.5) [Lp;Sp;Lp] (fun t u v->[t;u;v]);
      if ta.(1) then dessinesurf def dessin r true (0.5,0.5,0.5) [Lp;Sp;Lp] (fun t u v->[0.;u;v]);
      if ta.(1) then dessinesurf def dessin r true (0.5,0.5,0.5) [Lp;Sp;Lp] (fun t u v->[t;u;0.]);
      if ta.(2) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Sp;Rp] (fun t u v->[t;u;v]);
      if ta.(2) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Sp;Rp] (fun t u v->[0.;u;v]);
      if ta.(2) then dessinesurf def dessin r false (0.5,0.5,0.5) [Lp;Sp;Rp] (fun t u v->[t;u;0.]);
    end
